#include "_Plugin_Helper.h"
#ifdef USES_P003

// #######################################################################################################
// #################################### Plugin 003: Pulse  ###############################################
// #######################################################################################################
// KP: enhanced version with new PULSE Mode types 

#define P003_PULSE_STATISTIC        // additional debug/tuning messages from PULSE mode into the logfile

#define PLUGIN_003
#define PLUGIN_ID_003         3
#define PLUGIN_NAME_003       "Generic - Pulse counter"
#define PLUGIN_NR_VALUENAMES_003 3         // number of valuenames shown in Devices dialog
#define PLUGIN_VALUENAME1_003 "Count"
#define PLUGIN_VALUENAME2_003 "Total"
#define PLUGIN_VALUENAME3_003 "Time"
// their index in UserVar and TaskDeviceValueNames
#define P003_IDX_pulseCounter           0
#define P003_IDX_pulseTotalCounter      1
#define P003_IDX_pulseTime              2
#define P003_IDX_persistedTotalCounter  3  // This not shown variable is used to persist TotalCounter


// indexes for config parameters
#define P003_DEBOUNCETIME 0 
#define P003_COUNTERTYPE  1
#define P003_MODETYPE     2
// values for Mode types
#define P003_NR_MODETYPES       7
// special Mode Type. Note: only lower 3 bits are significant for GPIO Interupt type. Upper bits are flag to differentiate new modes
// for PULSE modes, we allways use the change trigger
#define PULSE_LOW    (0x10|CHANGE)
#define PULSE_HIGH   (0x20|CHANGE)
#define PULSE_CHANGE (0x30|CHANGE)
#define PULSE_MODE_MASK     0x30
#define MODE_INTERRUPT_MASK 0x03

// values for WEBFORM Counter Types
#define P003_NR_COUNTERTYPES    4
#define P003_COUNTERTYPE_LIST { F("Delta"), F("Delta/Total/Time"), F("Total"), F("Delta/Total"), }
#define P003_CT_INDEX_COUNTER              0
#define P003_CT_INDEX_COUNTER_TOTAL_TIME   1
#define P003_CT_INDEX_TOTAL                2
#define P003_CT_INDEX_COUNTER_TOTAL        3
// processing Steps in PLUGIN_TIMER_IN
#define P003_PROCESSING_STEP_1    1
#define P003_PROCESSING_STEP_2    2
#define P003_PROCESSING_STEP_3    3

void Plugin_003_pulse_interrupt1() ICACHE_RAM_ATTR;
void Plugin_003_pulse_interrupt2() ICACHE_RAM_ATTR;
void Plugin_003_pulse_interrupt3() ICACHE_RAM_ATTR;
void Plugin_003_pulse_interrupt4() ICACHE_RAM_ATTR;
void Plugin_003_pulsecheck(byte Index) ICACHE_RAM_ATTR;

// this takes 20 bytes of IRAM per handler
// void Plugin_003_pulse_interrupt5() ICACHE_RAM_ATTR;
// void Plugin_003_pulse_interrupt6() ICACHE_RAM_ATTR;
// void Plugin_003_pulse_interrupt7() ICACHE_RAM_ATTR;
// void Plugin_003_pulse_interrupt8() ICACHE_RAM_ATTR;

// main counters variables (used in ISR -> volatile)
volatile unsigned long P003_pulseCounter[TASKS_MAX];      // number of counted pulses within most recent data collection/sent interval 
volatile unsigned long P003_pulseTotalCounter[TASKS_MAX]; // total number of pulses counted since last reset
volatile unsigned long P003_pulseTime[TASKS_MAX];         // time between previous and most recent counted signal/pulse

// internal variables for PULSE mode
volatile bool P003_ProcessingFlag[TASKS_MAX];             // indicates that we, after an interrupt, are processing a pulse detection
unsigned long P003_pulseLowTime[TASKS_MAX];               // indicates the length in ms for the most recent stable low pulse
unsigned long P003_pulseHighTime[TASKS_MAX];              // indicates the length in ms for the most recent stable high pulse
byte          P003_currentStableState[TASKS_MAX];         // stores current stable pin state. Set in Step 3 when new stable pulse started
unsigned long P003_currentStableStartTime[TASKS_MAX];     // stores the start time of the current stable pulse. Set in Step 3
byte          P003_lastCheckState[TASKS_MAX];             // most recent pin state, that was read. Set in Step1,2,3

#ifdef P003_PULSE_STATISTIC
// debug/tuning variables for PULSE mode
volatile unsigned int P003_Step0counter[TASKS_MAX]; // counts how often step 0 was entered (volatile <- in ISR)
unsigned int P003_Step1counter[TASKS_MAX];          // counts how often step 1 was entered
unsigned int P003_Step2OKcounter[TASKS_MAX];        // counts how often step 2 detected the expected pin state (first verification)
unsigned int P003_Step2NOKcounter[TASKS_MAX];       // counts how often step 2 detected the wrong pin state (first verification failed)
unsigned int P003_Step3OKcounter[TASKS_MAX];        // counts how often step 3 detected the expected pin state (2nd verification)
unsigned int P003_Step3NOKcounter[TASKS_MAX];       // counts how often step 3 detected the wrong pin state (2nd verification failed)
unsigned int P003_Step3IGNcounter[TASKS_MAX];       // counts how often step 3 detected the wrong pin state (2nd verification failed)
#endif // P003_PULSE_STATISTIC


boolean Plugin_003(byte function, struct EventStruct *event, String& string)
{
  boolean success = false;
  String log = F("P003");

  switch (function)
  {
    case PLUGIN_DEVICE_ADD:   /** Set my technical parameters as ESPEasy device **/
    {
      Device[++deviceCount].Number           = PLUGIN_ID_003;
      Device[deviceCount].Type               = DEVICE_TYPE_SINGLE;
      Device[deviceCount].VType              = Sensor_VType::SENSOR_TYPE_SINGLE;
      Device[deviceCount].Ports              = 0;
      Device[deviceCount].PullUpOption       = false;
      Device[deviceCount].InverseLogicOption = false;
      Device[deviceCount].FormulaOption      = true;
      Device[deviceCount].ValueCount         = PLUGIN_NR_VALUENAMES_003;
      Device[deviceCount].SendDataOption     = true;
      Device[deviceCount].TimerOption        = true;
      Device[deviceCount].GlobalSyncOption   = true;
      break;
    }

    case PLUGIN_GET_DEVICENAME:   /** deliver my own device name **/
    {
      string = F(PLUGIN_NAME_003);
      break;
    }

    case PLUGIN_GET_DEVICEVALUENAMES:   /** define the names for the data that the plugin provides **/
    {
      strcpy_P(ExtraTaskSettings.TaskDeviceValueNames[0], PSTR(PLUGIN_VALUENAME1_003));
      strcpy_P(ExtraTaskSettings.TaskDeviceValueNames[1], PSTR(PLUGIN_VALUENAME2_003));
      strcpy_P(ExtraTaskSettings.TaskDeviceValueNames[2], PSTR(PLUGIN_VALUENAME3_003));
      break;
    }

    case PLUGIN_GET_DEVICEGPIONAMES:   /** return the pin name as show in the configuration webform **/
    {
      event->String1 = formatGpioName_input(F("Pulse"));
      break;
    }

    case PLUGIN_WEBFORM_LOAD:   /** add configuration entry boxes into webform html-code **/
    {
      addFormNumericBox(F("Debounce Time (mSec)"), F("p003")
                        , PCONFIG(P003_DEBOUNCETIME));

      byte   choice     = PCONFIG(P003_COUNTERTYPE);
      byte   choice2    = PCONFIG(P003_MODETYPE);
      String options[P003_NR_COUNTERTYPES] = P003_COUNTERTYPE_LIST;
      addFormSelector(F("Counter Type"), F("p003_countertype"), P003_NR_COUNTERTYPES, options, NULL, choice);

      /* KP: this problem was solved by using 4th UserVar for persistance. Thus this can be taken out
      if (choice != 0) {
        addHtml(F("<span style=\"color:red\">Total count is not persistent!</span>"));
      }
      */
      String modeRaise[P003_NR_MODETYPES];   // displayed texts
      // KP: correction for modeRaise[0]: LOW=0 did not generate interupts and does not make sense. Thus it changed to "none"
      //      Note: A correction to ONLOW = 0x04 (cf. Arduino.h) causes problems as it fires consecutive interupts, while GPIO is low 
      modeRaise[0] = F("none");
      modeRaise[1] = F("CHANGE");
      modeRaise[2] = F("RISING");
      modeRaise[3] = F("FALLING");
      modeRaise[4] = F("PULSE change");
      modeRaise[5] = F("PULSE high");
      modeRaise[6] = F("PULSE low");
      
      int modeValues[P003_NR_MODETYPES];  // trigger flags
      modeValues[0] = 0;    // KP: replaced LOW by 0, as it does nothing
      modeValues[1] = CHANGE;
      modeValues[2] = RISING;
      modeValues[3] = FALLING;
      modeValues[4] = PULSE_CHANGE;;
      modeValues[5] = PULSE_HIGH;;
      modeValues[6] = PULSE_LOW;;

      addFormSelector(F("Mode Type"), F("p003_raisetype"), P003_NR_MODETYPES, modeRaise, modeValues, choice2);

      success = true;
      break;
    }

    case PLUGIN_WEBFORM_SAVE:   /** save configuration date as provided by webform **/
    {
      PCONFIG(P003_DEBOUNCETIME)  = getFormItemInt(F("p003"));
      PCONFIG(P003_COUNTERTYPE)   = getFormItemInt(F("p003_countertype"));
      PCONFIG(P003_MODETYPE)      = getFormItemInt(F("p003_raisetype"));
      success                     = true;
      break;
    }

    case PLUGIN_WEBFORM_SHOW_VALUES:   /** provide current plugin data to webform  **/
    {
      pluginWebformShowValue(ExtraTaskSettings.TaskDeviceValueNames[P003_IDX_pulseCounter],      String(P003_pulseCounter[event->TaskIndex]));
      pluginWebformShowValue(ExtraTaskSettings.TaskDeviceValueNames[P003_IDX_pulseTotalCounter], String(P003_pulseTotalCounter[event->TaskIndex]));
      pluginWebformShowValue(ExtraTaskSettings.TaskDeviceValueNames[P003_IDX_pulseTime],         String(P003_pulseTime[event->TaskIndex]), false);
      success = true;
      break;
    }

    case PLUGIN_INIT:   /** initialize plugin **/
    {
      // Restore any value that may have been read from the RTC.
      switch (PCONFIG(P003_COUNTERTYPE))
      {
        case P003_CT_INDEX_COUNTER:
        {
          P003_pulseCounter[event->TaskIndex] = UserVar[event->BaseVarIndex + P003_IDX_pulseCounter];
          break;
        }
        case P003_CT_INDEX_COUNTER_TOTAL_TIME:
        {
          P003_pulseCounter[event->TaskIndex]      = UserVar[event->BaseVarIndex + P003_IDX_pulseCounter];
          P003_pulseTotalCounter[event->TaskIndex] = UserVar[event->BaseVarIndex + P003_IDX_pulseTotalCounter];
          P003_pulseTime[event->TaskIndex]         = UserVar[event->BaseVarIndex + P003_IDX_pulseTime];
          break;
        }
        case P003_CT_INDEX_TOTAL:
        {
          P003_pulseTotalCounter[event->TaskIndex] = UserVar[event->BaseVarIndex + P003_IDX_pulseCounter];
          break;
        }
        case P003_CT_INDEX_COUNTER_TOTAL:
        {
          P003_pulseCounter[event->TaskIndex]      = UserVar[event->BaseVarIndex + P003_IDX_pulseCounter];
          P003_pulseTotalCounter[event->TaskIndex] = UserVar[event->BaseVarIndex + P003_IDX_pulseTotalCounter];
          break;
        }
      }

      // Restore the total counter from the unused 4th UserVar value.
      // It may be using a formula to generate the output, which makes it impossible to restore
      // the true internal state.
      P003_pulseTotalCounter[event->TaskIndex] = UserVar[event->BaseVarIndex + P003_IDX_persistedTotalCounter];

      // initialize internal variables for PULSE mode handling
      P003_ProcessingFlag[event->TaskIndex]         = false;
      P003_pulseLowTime[event->TaskIndex]           = 0;
      P003_pulseHighTime[event->TaskIndex]          = 0;
      P003_currentStableState[event->TaskIndex]     = HIGH;
      P003_currentStableStartTime[event->TaskIndex] = 0;
      P003_lastCheckState[event->TaskIndex]         = HIGH;    

      #ifdef P003_PULSE_STATISTIC
      // initialize statistical step counters from TotalCounter and error counters with 0
      P003_Step0counter[event->TaskIndex]     = P003_pulseTotalCounter[event->TaskIndex];
      P003_Step1counter[event->TaskIndex]     = P003_pulseTotalCounter[event->TaskIndex];
      P003_Step2OKcounter[event->TaskIndex]   = P003_pulseTotalCounter[event->TaskIndex];
      P003_Step2NOKcounter[event->TaskIndex]  = 0;
      P003_Step3OKcounter[event->TaskIndex]   = P003_pulseTotalCounter[event->TaskIndex];
      P003_Step3NOKcounter[event->TaskIndex]  = 0;
      P003_Step3IGNcounter[event->TaskIndex]  = 0;
      #endif // P003_PULSE_STATISTIC

      log = F("INIT : PulsePin: ");
      log += Settings.TaskDevicePin1[event->TaskIndex];
      addLog(LOG_LEVEL_INFO, log);
      
      // set up device pin and estabish interupt handlers
      pinMode(Settings.TaskDevicePin1[event->TaskIndex], INPUT_PULLUP);
      success =
        Plugin_003_pulseinit(Settings.TaskDevicePin1[event->TaskIndex], event->TaskIndex,
                             (PCONFIG(P003_MODETYPE) & MODE_INTERRUPT_MASK));

      break;
    }

    case PLUGIN_READ:   /** return current plugin data into UserVar **/
    {
      // FIXME TD-er: Is it correct to write the first 3  UserVar values, regardless the set counter type?
      UserVar[event->BaseVarIndex + P003_IDX_pulseCounter]      = P003_pulseCounter[event->TaskIndex];
      UserVar[event->BaseVarIndex + P003_IDX_pulseTotalCounter] = P003_pulseTotalCounter[event->TaskIndex];
      UserVar[event->BaseVarIndex + P003_IDX_pulseTime]         = P003_pulseTime[event->TaskIndex];

      // Store the raw value in the unused 4th position.
      // This is needed to restore the value from RTC as it may be converted into another output value using a formula.
      UserVar[event->BaseVarIndex + P003_IDX_persistedTotalCounter] = P003_pulseTotalCounter[event->TaskIndex];


      switch (PCONFIG(P003_COUNTERTYPE))
      {
        case P003_CT_INDEX_COUNTER:
        {
          event->sensorType = Sensor_VType::SENSOR_TYPE_SINGLE;
          UserVar[event->BaseVarIndex + P003_IDX_pulseCounter] = P003_pulseCounter[event->TaskIndex];
          break;
        }
        case P003_CT_INDEX_COUNTER_TOTAL_TIME:
        {
          event->sensorType = Sensor_VType::SENSOR_TYPE_TRIPLE;
          UserVar[event->BaseVarIndex + P003_IDX_pulseCounter]      = P003_pulseCounter[event->TaskIndex];
          UserVar[event->BaseVarIndex + P003_IDX_pulseTotalCounter] = P003_pulseTotalCounter[event->TaskIndex];
          UserVar[event->BaseVarIndex + P003_IDX_pulseTime]         = P003_pulseTime[event->TaskIndex];
          break;
        }
        case P003_CT_INDEX_TOTAL:
        {
          event->sensorType = Sensor_VType::SENSOR_TYPE_SINGLE;
          UserVar[event->BaseVarIndex + P003_IDX_pulseCounter]     = P003_pulseTotalCounter[event->TaskIndex];
          break;
        }
        case P003_CT_INDEX_COUNTER_TOTAL:
        {
          event->sensorType = Sensor_VType::SENSOR_TYPE_DUAL;
          UserVar[event->BaseVarIndex + P003_IDX_pulseCounter]      = P003_pulseCounter[event->TaskIndex];
          UserVar[event->BaseVarIndex + P003_IDX_pulseTotalCounter] = P003_pulseTotalCounter[event->TaskIndex];
          break;
        }
      }
      P003_pulseCounter[event->TaskIndex] = 0;
      success                             = true;
      break;
    }

    case PLUGIN_WRITE:   /** execute functional command **/
    {
      String command            = parseString(string, 1);
      bool   mustCallPluginRead = false;

      if (command == F("resetpulsecounter"))
      {
        // Valid commands:
        // - resetpulsecounter
        // - resetpulsecounter,taskindex

        // Allow for an optional taskIndex parameter. When not given it will take the first task with this plugin.
        if (!pluginOptionalTaskIndexArgumentMatch(event->TaskIndex, string, 1)) {
          break;
        }

        #ifdef P003_PULSE_STATISTIC
        // reduce the statistical step counters by the current value of TotalCounter in order to keep statistic correct
        P003_Step0counter[event->TaskIndex]     -= P003_pulseTotalCounter[event->TaskIndex];
        P003_Step1counter[event->TaskIndex]     -= P003_pulseTotalCounter[event->TaskIndex];
        P003_Step2OKcounter[event->TaskIndex]   -= P003_pulseTotalCounter[event->TaskIndex];
        P003_Step3OKcounter[event->TaskIndex]   -= P003_pulseTotalCounter[event->TaskIndex];
        #endif // P003_PULSE_STATISTIC

        P003_pulseCounter[event->TaskIndex]      = 0;
        P003_pulseTotalCounter[event->TaskIndex] = 0;
        P003_pulseTime[event->TaskIndex]         = 0;

        mustCallPluginRead                       = true;
        success                                  = true; // Command is handled.
      } 
      else if (command == F("setpulsecountertotal"))
      {
        // Valid commands:
        // - setpulsecountertotal,value
        // - setpulsecountertotal,value,taskindex

        // First check if (optional) task index matches.
        if (!pluginOptionalTaskIndexArgumentMatch(event->TaskIndex, string, 2)) {
          break;
        }

        int par1;

        if (validIntFromString(parseString(string, 2), par1))
        {
          #ifdef P003_PULSE_STATISTIC
          // adjust the statistical step counters relative to TotalCounter, in order to keep statistic correct
          P003_Step0counter[event->TaskIndex]     -= P003_pulseTotalCounter[event->TaskIndex] - par1;
          P003_Step1counter[event->TaskIndex]     -= P003_pulseTotalCounter[event->TaskIndex] - par1;
          P003_Step2OKcounter[event->TaskIndex]   -= P003_pulseTotalCounter[event->TaskIndex] - par1;
          P003_Step3OKcounter[event->TaskIndex]   -= P003_pulseTotalCounter[event->TaskIndex] - par1;
          #endif // P003_PULSE_STATISTIC

          P003_pulseTotalCounter[event->TaskIndex] = par1;

          mustCallPluginRead                       = true;
          success                                  = true; // Command is handled.
        }
      }

      if (mustCallPluginRead) {
        // Note that the set time is before the current moment, so we call the read as soon as possible.
        // The read does also use any set formula and stored the value in RTC.
        Scheduler.schedule_task_device_timer(event->TaskIndex, millis() - 10);
      }
      break;
    }

    case PLUGIN_TIMER_IN:   /** process timer event **/
    {
      // this function is called when the next processing step is scheduled
      int pinState;
      int processingStep = event->Par1;

      switch (processingStep)
      {
        case P003_PROCESSING_STEP_1:   // read pin status
        {
          #ifdef P003_PULSE_STATISTIC
          P003_Step1counter[event->TaskIndex]++;
          #endif // P003_PULSE_STATISTIC
          //  read current state from this tasks's GPIO
          P003_lastCheckState[event->TaskIndex] = (byte)digitalRead(Settings.TaskDevicePin1[event->TaskIndex]);
          // after debounceTime/2, do step 2          
          Scheduler.setPluginTaskTimer(PCONFIG(P003_DEBOUNCETIME) >> 1, event->TaskIndex, P003_PROCESSING_STEP_2);
          break;  
        }
        
        case P003_PROCESSING_STEP_2:   // 1st validation of pin status
        {
          //  read current state from this tasks's GPIO
          pinState = digitalRead(Settings.TaskDevicePin1[event->TaskIndex]);

          if (pinState == (int)P003_lastCheckState[event->TaskIndex] )
          // we found stable state
          {
            #ifdef P003_PULSE_STATISTIC
            P003_Step2OKcounter[event->TaskIndex]++;
            #endif // P003_PULSE_STATISTIC
            // after debounceTime/2, do step 3          
            Scheduler.setPluginTaskTimer(PCONFIG(P003_DEBOUNCETIME >> 1), event->TaskIndex, P003_PROCESSING_STEP_3);
          }
          else
          // we found unexpected different pin state
          {
            #ifdef P003_PULSE_STATISTIC
            P003_Step2NOKcounter[event->TaskIndex]++;
            #endif // P003_PULSE_STATISTIC
            // lets ignore previous pin status. It might have been a pieks. Try to detect stable signal
            P003_lastCheckState [event->TaskIndex] = pinState;    // now trust the new state
            // after debounceTime/2, do step 2 again     
            Scheduler.setPluginTaskTimer(PCONFIG(P003_DEBOUNCETIME >> 1), event->TaskIndex, P003_PROCESSING_STEP_2);

          }      
          break;  
        }

        case P003_PROCESSING_STEP_3:  // 2nd validation of pin status
        {
          //  read current state from this tasks's GPIO
          pinState = digitalRead(Settings.TaskDevicePin1[event->TaskIndex]);

          if (pinState == (int)P003_lastCheckState[event->TaskIndex] )
          // we found stable state
          {

            if (pinState != (int)P003_currentStableState[event->TaskIndex] )
            // we found new stable state
            {
              #ifdef P003_PULSE_STATISTIC
              P003_Step3OKcounter[event->TaskIndex]++;
              #endif // P003_PULSE_STATISTIC
              // effective start time of pulse is NOW - 2 * DebounceTime
              unsigned long pulseChangeTime = millis() - (PCONFIG(P003_DEBOUNCETIME) << 1);

              // lets terminate the previous pulse and setup start point for current stable one
              if ( P003_currentStableState[event->TaskIndex] == HIGH )  // HIGH has ended
                { P003_pulseHighTime[event->TaskIndex] = pulseChangeTime - P003_currentStableStartTime[event->TaskIndex]; }
              else    // LOW has ended
                { P003_pulseLowTime[event->TaskIndex]  = pulseChangeTime - P003_currentStableStartTime[event->TaskIndex]; }  
              P003_currentStableState[event->TaskIndex]     = !P003_currentStableState[event->TaskIndex];
              P003_currentStableStartTime[event->TaskIndex] = pulseChangeTime;       


              // now provide the counter result values for the ended pulse ( depending on mode type)
              switch (Settings.TaskDevicePluginConfig[event->TaskIndex][P003_MODETYPE])
              {
                case PULSE_CHANGE:
                {
                  if (P003_currentStableState[event->TaskIndex] == LOW )  // HIGH had ended
                    P003_pulseTime[event->TaskIndex] = P003_pulseHighTime[event->TaskIndex];
                  else    // LOW has ended
                    P003_pulseTime[event->TaskIndex] = P003_pulseLowTime[event->TaskIndex];

                  P003_pulseCounter[event->TaskIndex]++;
                  P003_pulseTotalCounter[event->TaskIndex]++;
                  break;
                }
                case PULSE_HIGH:
                {
                  if ( P003_currentStableState[event->TaskIndex] == LOW )  // HIGH had ended (else do nothing)
                  {
                    P003_pulseTime[event->TaskIndex] = P003_pulseLowTime[event->TaskIndex] + P003_pulseHighTime[event->TaskIndex];
                    P003_pulseCounter[event->TaskIndex]++;
                    P003_pulseTotalCounter[event->TaskIndex]++;
                  }
                  break;
                }
                case PULSE_LOW:
                {
                  if ( P003_currentStableState[event->TaskIndex] == HIGH )  // LOW had ended (else do nothing)
                  {
                    P003_pulseTime[event->TaskIndex] = P003_pulseLowTime[event->TaskIndex] + P003_pulseHighTime[event->TaskIndex];
                    P003_pulseCounter[event->TaskIndex]++;
                    P003_pulseTotalCounter[event->TaskIndex]++;
                  }
                  break;
                }
                default:
                {
                  log = F("_P003:PLUGIN_TIMER_IN: Invalid modeType: ");
                  log += Settings.TaskDevicePluginConfig[event->TaskIndex][P003_MODETYPE];
                  addLog(LOG_LEVEL_ERROR, log);
                  break;
                }
              }
            }
            else 
            // we found the same stable state as before
            {
              #ifdef P003_PULSE_STATISTIC
              P003_Step3IGNcounter[event->TaskIndex]++;
              #endif // P003_PULSE_STATISTIC
              // do nothing. Ignore interupt. previous stable state was confirmed probably after a pieks
            }

            #ifdef P003_PULSE_STATISTIC
            // Statistic to logfile. E.g: ... [123 | 111 | 100/5 | 80/3/4/ | 40 ][ 12243 | 3244 ]
            log = F("_P003: Pulse Stats [step 0|1|2|3|total (ok/nok/ign)][low ms|high ms]: [ ");
            log += P003_Step0counter[event->TaskIndex];      log += " | ";
            log += P003_Step1counter[event->TaskIndex];      log += " | ";
            log += P003_Step2OKcounter[event->TaskIndex];    log += "/";
            log += P003_Step2NOKcounter[event->TaskIndex];   log += " | ";
            log += P003_Step3OKcounter[event->TaskIndex];    log += "/";
            log += P003_Step3NOKcounter[event->TaskIndex];   log += "/";
            log += P003_Step3IGNcounter[event->TaskIndex];   log += " | ";
            log += P003_pulseTotalCounter[event->TaskIndex]; log += " ][ ";
            log += P003_pulseLowTime[event->TaskIndex];      log += " | ";
            log += P003_pulseHighTime[event->TaskIndex];     log += " ] ";
            addLog(LOG_LEVEL_INFO, log);
            #endif // P003_PULSE_STATISTIC

            // allow further pulse interrupts
            P003_ProcessingFlag[event->TaskIndex] = false;
          }
          else
          // we found unexpected different pin state
          {
            #ifdef P003_PULSE_STATISTIC
            P003_Step3NOKcounter[event->TaskIndex]++;
            #endif // P003_PULSE_STATISTIC
            // ignore piek from previous step. Treated as piek within previous=current signal. Try to detect stable signal again
            P003_lastCheckState [event->TaskIndex] = pinState;    // now trust the new state
            Scheduler.setPluginTaskTimer(PCONFIG(P003_DEBOUNCETIME) >> 1, event->TaskIndex, P003_PROCESSING_STEP_2);
          }
          break;  
        }
        default:
        {
          log = F("_P003:PLUGIN_TIMER_IN: Invalid processingStep: ");
          log += processingStep;
          addLog(LOG_LEVEL_ERROR, log);
          break;
        }
      }
      break;
    }

  }
  return success;
}

/*********************************************************************************************\
* Check Pulse Counters (called from irq handler)
\*********************************************************************************************/
void Plugin_003_pulsecheck(byte Index)
{
  
  if ( (Settings.TaskDevicePluginConfig[Index][2] & PULSE_MODE_MASK) == 0 )
  // legathy Mode types
  {
    /* KP: Remark: I wonder what the intention was for using "noInterrupts()"
    because the interupt is automatically disabled by ESP8266 while we are in the ISR
    and it thus does not help to reduce bouncing here. What problem was intended to solve?
    It is furthermore generally riscy to incondicionaly use "interrupts()" (rather than restore the previos state)
    because it could disturb higher-level functions, when we unexpectedly enable interrupts in lower-level functions.
    I wonder whether we can better leave the interrupt handling out here. 
    */
    noInterrupts(); // s0170071: avoid nested interrups due to bouncing.
    //  s0170071: the following gives a glitch if millis() rolls over (every 50 days) and there is a bouncing to be avoided at the exact same
    // time. Very rare.
    //  Alternatively there is timePassedSince(P003_pulseTimePrevious[Index]); but this is not in IRAM at this time, so do not use in a
    // ISR!
    // KP: we use here P003_currentStableStartTime[Index] from new Pulse modes to persist the PulseTime.
    const unsigned long timeSinceLastTrigger = millis() - P003_currentStableStartTime[Index];

    if (timeSinceLastTrigger > (unsigned long)Settings.TaskDevicePluginConfig[Index][P003_DEBOUNCETIME]) // check with debounce time for this task
    {
      P003_pulseCounter[Index]++;
      P003_pulseTotalCounter[Index]++;
      P003_pulseTime[Index]               = timeSinceLastTrigger;
      P003_currentStableStartTime[Index]  = millis();
    }
    interrupts(); // enable interrupts again.
  }
  else
  // processing for new PULSE mode tpes
  {

    #ifdef P003_PULSE_STATISTIC
    P003_Step0counter[Index]++;
    #endif // P003_PULSE_STATISTIC
    
    if ( !P003_ProcessingFlag[Index] )  // only process interrupt if we are not still processing the previous one
    {
      P003_ProcessingFlag[Index] = true;  // indicate that we are now processing the interrupt      
      // schedule processing step 1 after debounce time ellapsed
      Scheduler.setPluginTaskTimer(Settings.TaskDevicePluginConfig[Index][P003_DEBOUNCETIME], Index, P003_PROCESSING_STEP_1);
    }
  }

}

/*********************************************************************************************\
* Pulse Counter IRQ handlers
\*********************************************************************************************/
void Plugin_003_pulse_interrupt1()
{
  Plugin_003_pulsecheck(0);
}

void Plugin_003_pulse_interrupt2()
{
  Plugin_003_pulsecheck(1);
}

void Plugin_003_pulse_interrupt3()
{
  Plugin_003_pulsecheck(2);
}

void Plugin_003_pulse_interrupt4()
{
  Plugin_003_pulsecheck(3);
}

void Plugin_003_pulse_interrupt5()
{
  Plugin_003_pulsecheck(4);
}

void Plugin_003_pulse_interrupt6()
{
  Plugin_003_pulsecheck(5);
}

void Plugin_003_pulse_interrupt7()
{
  Plugin_003_pulsecheck(6);
}

void Plugin_003_pulse_interrupt8()
{
  Plugin_003_pulsecheck(7);
}

/*********************************************************************************************\
* Init Pulse Counters
\*********************************************************************************************/
bool Plugin_003_pulseinit(byte Par1, byte Index, byte Mode)
{
  switch (Index)
  {
    case 0:
      attachInterrupt(Par1, Plugin_003_pulse_interrupt1, Mode);
      break;
    case 1:
      attachInterrupt(Par1, Plugin_003_pulse_interrupt2, Mode);
      break;
    case 2:
      attachInterrupt(Par1, Plugin_003_pulse_interrupt3, Mode);
      break;
    case 3:
      attachInterrupt(Par1, Plugin_003_pulse_interrupt4, Mode);
      break;

    // case 4:
    //   attachInterrupt(Par1, Plugin_003_pulse_interrupt5, Mode);
    //   break;
    // case 5:
    //   attachInterrupt(Par1, Plugin_003_pulse_interrupt6, Mode);
    //   break;
    // case 6:
    //   attachInterrupt(Par1, Plugin_003_pulse_interrupt7, Mode);
    //   break;
    // case 7:
    //   attachInterrupt(Par1, Plugin_003_pulse_interrupt8, Mode);
    //   break;
    default:
      addLog(LOG_LEVEL_ERROR, F("PULSE: Error, only the first 4 tasks can be pulse counters."));
      return false;
  }

  return true;
}

#endif // USES_P003
